#perl
#
# Title: Export Source Engine .smd Mesh
# Version: 1.2
# Author: James O'Hare
# http://www.farfarer.com
# jamesohare@gmail.com
#

#
# Description:
# Exports a mesh to an .smd file for use in games running on the Source
# engine (Half-Life2, Counter-Strike: Source, Team Fortress 2, Day of
# Defeat, # etc...).
#
# Excuse the bad code, I'm new to this programming stuff :)
#

#
# Usage:
# Currently it will only export your mesh with 1 UV set and the texture
# name it exports is the filename of the first diffuse texture applied
# to the material of that given polygon. Defaults to default.tga if none
# are applied.
#
# It will export either the UV map you have selected or the first one it finds.
# It will export all visisble polygons in all active layers.
#
# Whatever units you are using, 1 of them will equal 1 inch in your exported model.
# e.g. A character in modo should be 72 [your scale units] high, meaning the mesh
# will be 72 inches (6 foot tall) in-game.
# Coincidentally, 1 inch on an in-game model is exactly the same as 1 unit in Hammer.
#

#
# NOTE:
# When you export a mesh with several thousand tris, it can end up taking a very long
# time. It may appear that modo has crashed, but it hasn't - it's just working very
# hard. Give it a few minutes and it'll come back. I will try and make this faster in
# the future.
#

#
# To-do list for future versions:
# - Basic animation (per-object).
# - Morphs for .VTA morph maps.
# - Smoothing based of material smoothing angle.
#

#
# If you find the script breaks on certain meshes, you can email me the mesh at
# jamesohare@gmail.com and I'll try and work out what's wrong.
#

#======================================================================================== Common Stuff =

#Let's set a hard limit of 25k tris.
sub checkTris {
	my $numPolys = lxq("query layerservice poly.N ? all");
	if ($numPolys > 25000) {
		lxout("$numPolys tris");
		$excess = $numPolys - 25000;
		die(" \n \nWARNING!\n \nYou have $excess too many triangles ($numPolys in total). The maximum is 25,000 triangles.\n \n ");
	}
}

# Because we're copying everything to a new scene, we have to check what UV map is selected now and then pass that info through to the next function.
sub getUV {
	my @vmapSelected = lxq("query layerservice vmaps ? selected");

	foreach $vmap (@vmapSelected) {
		my $vmapType = lxq("query layerservice vmap.type ? $vmap");
		if ($vmapType eq "texture") {
			$vmapID = $vmap;
			last;
		}
	}

	return $vmapID;
}

# Copy everything to a new scene, paste it and triple it. I don't want to screw over the original file if this goes wrong by accident.
sub newScene {
	lx("select.type item");
	lx("select.type polygon");
	lx("select.drop polygon");
	lx("select.copy");
	lx("scene.new");
	lx("select.paste");
	#This has to be here as the script was last querying vertex maps and so it won't triple stuff unless we go back and query some meshes.
	my $mainlayer = lxq("query layerservice layer.id ? main");
	lx("poly.triple");
}

# Close the temp scene I created and suppress it's warnings.
sub endScene {
	lx("!!scene.close");
}

# Seems like SMD files only uses 6 decimal places. Modo spits out much more than that, so we have to cut it down.
# It has a convenient side-effect of setting anything that gets put through it to 0 if it wasn't a number in the first place.
sub round {
	return sprintf("%.6f", $_[0]);
}

# Takes a material mask in and returns the first diffuse colour image map in it, renamed to a TGA.
sub imageMap {
	# Get the children of the mask (all it's layers and materials).
	my @txChildren = lxq("query sceneservice txLayer.children ? $_[0]");
	
	#Cycle through all the children and the first diffColor texture, then get the clip name and return it.
	foreach $txChild (@txChildren) {
		lx("select.subItem $txChild set textureLayer;render;environment");
		my $channel = lxq("shader.setEffect ?");
		if ($channel eq "diffColor") {
			my $IMap = lxq("texture.setIMap ?");
			my $fileName = $clipHash{$IMap};
			my $fileName = fixFileName($fileName);
			$childHash{$_[1]} = $fileName;
			return $fileName;
			lxout("texture: $fileName");
			last;
		}
	}
}

# Strip out the directory and rename the file to a TGA.
sub fixFileName {
	my @splitFullName = split /\\/, $_[0];
	my $fileName = @splitFullName[-1];
	# If it doesn't have the extension tga, strip the extension and make it .tga
	# This means you can use .PSD files on your mesh, export it and it'll still
	# look for the right materials when compiled.
	if($fileName !~ m/\.[Tt][Gg][Aa]/)
	{
		$fileName =~ s/(\.[^.]+)$//;
		$fileName = "$fileName.tga";
	}
	return $fileName;
}

# Create some commonly used hashes - these help speed the processing up by storing lookup values that would otherwise require thousands of iterative queries.
sub makeHashes {
	my %textureToClip = ();

	my @texs = lxq("query layerservice textures ?");
	foreach $tex (@texs) {
		my $texType = lxq("query layerservice texture.type ? $tex");

		if ( $texType eq "imageMap" ) {
			my $texLocator = lxq("query layerservice texture.locator ? $tex");
			my $texName = lxq("query layerservice texture.name ? $tex");
			my $texID = lxq("query layerservice texture.id ? $tex");
			my $texUVName = lxq("query layerservice texture.uvName ? $tex");
			my $texClip = lxq("query layerservice texture.clip ? $tex");
			my $clipName = lxq("query layerservice clip.name ? $texClip");

			lx("select.subItem $texLocator set txtrLocator");
			lx("select.vertexMap $texUVName txuv replace");
			lx("select.subItem $clipName set mediaClip");
			lx("select.subItem $texID set textureLayer;render;environment;light;camera;scene;replicator");

			my $shaderEffect = lxq("shader.setEffect ?");

			#lxout("Name: $texName   Type: $texType   ID: $texID   Channel: $shaderEffect   ClipName: $clipName");

			if ( $shaderEffect eq "diffColor" ) {
				$textureToClip{$texID} = $clipName;
				break;
			}
		}
	}

	%materialToTexture = ();

	my @mats = lxq("query layerservice materials ?");
	foreach $mat (@mats) {
		my $matName = lxq("query layerservice material.name ? $mat");
		@matTexs = lxq("query layerservice material.textures ? $mat");
		foreach $matTex (@matTexs) {
			if (defined $textureToClip{$matTex}) {
				$materialToTexture{$matName} = $textureToClip{$matTex};
				break;
			}
			#lxout ("$matName   $matTex");
		}
	}
}

#======================================================================================== File Handling =

# Open the specified file for SMD export.
sub openFile
{
	lxout("opening $_[0]");
	open(SMD, ">", $_[0]) or die "Unable to open $_[0]";
}

# Write the file contents.
sub writeFile {
	lxout("writing file $_[0]");
	print SMD "$_[1]";
}

# Close the file.
sub closeFile {
	lxout("closing $_[0]");
	close (SMD);
}

# Save dialog.

sub getFileName {
	lx("dialog.setup fileSave");
	lx("dialog.title {Export to SMD}");
	lx("dialog.fileTypeCustom format:[sim] username:[Source reference SMD] loadPattern:[]  saveExtension:[smd]");
	lx("dialog.open");
	my $file = lxq("dialog.result ?");
	if (!$file) {
		die("User Cancelled Script");
	} else {	
		return $file;
	}
}

#======================================================================================== Data Gathering =

sub getContent {

	checkTris();
	makeHashes();

	# Get the currently selected vertex map and check that it's a UV map.
	# If it's not, select the first one.
	# If there is no first one, the script will zero all UV coords.
	my @vmapList = lxq("query layerservice vmaps ? all");
	my $vmapID = "";
	
	if (defined $_[0]) {
		$vmapID = $_[0];
	}
	else
	{
		foreach $vmap (@vmapList) {
			$vmapType = lxq("query layerservice vmap.type ? $vmap");
			if ($vmapType eq "texture") {
				$vmapID = $vmap;
				last;
			}
		}
	}

	my $vmapNormalID = "";
	
	foreach $vmap (@vmapList) {
		$vmapType = lxq("query layerservice vmap.type ? $vmap");
		if ($vmapType eq "normal") {
			$vmapNormalID = $vmap;
			last;
		}
	}
	
	my $smdContent = "version 1
nodes
 0 \"smdExport\"  -1
end
skeleton
time 0
 0 0.000000 0.000000 0.000000 0.000000 0.000000 0.000000
end
triangles";
	my @polyList = lxq("query layerservice polys ? all");
	my @texList = lxq("query layerservice textures ? all");
	foreach $poly (@polyList) {
		# Get the material applied to the polygon. We'll use this to deduce the image map applied to it.
		my $material = lxq("query layerservice poly.material ? $poly");
		#lxout("Material: $material");
		
		# Default image to apply. If an image is found applied to the polygon, this value will be overwritten.
		my $image = "default";

		# Cycle through the textures and find the one whose ID matches that of the tag applied to the poly.
		# This means that's the mask, which we can then pass off to retrieve it's first diffuse colour image map.
		
		if (defined $materialToTexture{$material}) {
			$image = $materialToTexture{$material};
		}
		$smdContent = "$smdContent\n$image.tga";

		# Get it's vertices and UV coord array (only if a UV map exists, otherwise it'll just spit out zeroes").
		if ($vmapID ne "") {
			my $selVMap = lxq("query layerservice vmap.name ? $vmapID");
			@uvCoords = lxq("query layerservice poly.vmapValue ? $poly");
		}
		# Round the values off.
		@uvCoords[0] = round(@uvCoords[0]);
		@uvCoords[1] = round(@uvCoords[1]);
		@uvCoords[2] = round(@uvCoords[2]);
		@uvCoords[3] = round(@uvCoords[3]);
		@uvCoords[4] = round(@uvCoords[4]);
		@uvCoords[5] = round(@uvCoords[5]);
		
		my @vertList = lxq("query layerservice poly.vertList ? $poly");
		my $vertCount = 0;
		
		
		# Check to see if we found a vertex normal map earlier and query it if we did - this means it'll use this vertex map when we query it later.
		if ($vmapNormalID ne "") {
			my $selVMap = lxq("query layerservice vmap.name ? $vmapNormalID");
		}
		
		foreach $vert (@vertList) {
			my @vertPos = lxq("query layerservice vert.pos ? $vert");
			# Round the values off.
			#Invert the Z coord here as it's really the Y coord in-game and it'll be backwards and inside out if it's not inverted.
			@vertPos[0] = round(@vertPos[0]);
			@vertPos[1] = round(@vertPos[1]);
			@vertPos[2] = round(@vertPos[2])*-1;
			
			my @vertNorm = ();
			if ($vmapNormalID ne "") {
				@vertNorm = lxq("query layerservice vert.vmapValue ? $vert");
			}
			else {
				@vertNorm = lxq("query layerservice vert.normal ? $vert");
			}
			# Round the values off.
			@vertNorm[0] = round(@vertNorm[0]);
			@vertNorm[1] = round(@vertNorm[1]);
			@vertNorm[2] = round(@vertNorm[2])*-1;

		#Swapped Y and Z here so that the model is upright in-game.
		
		# Format per vert is <Parent> <PosX> <PosY> <PosZ> <NormX> <NormY> <NormZ> <TexU> <TexV>
		$smdContent = "$smdContent\n 0   @vertPos[0]   @vertPos[2]   @vertPos[1]   @vertNorm[0]   @vertNorm[2]   @vertNorm[1]   @uvCoords[$vertCount*2]   @uvCoords[$vertCount*2+1]";
		
		$vertCount++;
		
		}
	}
	
	# The newline after the end here is important. studiomdl.exe will not recognise that the line has been terminated without one.
	$smdContent = "$smdContent \nend\n";

	return $smdContent;
}

#========================================================================================== File Writing =

# Create, open, write and close the file.

my $filename = getFileName();
openFile($filename);
my $selectedUV = getUV(); # Pull the index of the current UV map.
newScene();
my $contents = getContent($selectedUV); # Pass it through to the new scene to be selected.
writeFile($filename, $contents);
closeFile($filename);
endScene();